/*
 * Worldcup is Copyright (c) 2001,2002 by Thomas Pundt
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose, without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice and this paragraph and the
 * following two paragraphs appear in all copies.
 *
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 * THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
 * BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include <PalmOS.h>

#include "worldcup.h"
#include "worldcuprcp.h"

struct match match[8];

struct mstatus MStatus[4], Equal[4];

static UInt8 runden[5] = {0,8,12,14,16};

MyAppPreferencesType appPrefs;
UInt8 selectedStadium = 0;
UInt8 selectedTeam = 0;
Int32 defaultDateFormat;

static TablePtr leagueTable;

static void SetDatumString(Int16 id, UInt32 datum)
{
  DateTimeType dateP;
  FormType* frm = FrmGetActiveForm();
  char *label = (char*)FrmGetLabel(frm,id);

  FrmHideObject(frm,FrmGetObjectIndex(frm,id));
  TimSecondsToDateTime(datum+appPrefs.gmtOff, &dateP);
  DateToAscii(dateP.month,dateP.day,dateP.year,
	      PrefGetPreference(prefDateFormat), label);

  // don't draw 0:00 time
  if (dateP.hour+dateP.minute>0) {
    StrCat(label," ");
    TimeToAscii(dateP.hour,dateP.minute,tfColon24h,
		label+StrLen(label));
  }
  FrmShowObject(frm,FrmGetObjectIndex(frm,id));
}

static void SetRoundString(Int16 id, int num)
{
  char *p, *label;
  ControlType *ctlP = Id2Ptr(id);
  label = (char*)CtlGetLabel(ctlP);
  p = GetString(num);
  CtlEraseControl(ctlP);
  StrCopy(label,p);
  CtlDrawControl(ctlP);
}

static void SetTimezoneString(Int16 id, int num, const Char* format)
{
  char *label;
  ControlType *ctlP = Id2Ptr(id);
  label = (char*)CtlGetLabel(ctlP);
  StrPrintF(label,format,num);
  CtlEraseControl(ctlP);
  CtlDrawControl(ctlP);
}

/*
 * This comparison function takes into account items 6.a)-6.c) from
 * Article 28) of FIFA Worldcup 2002 Reglement or items 6.d)-6.g)
 * (points, goals and goals difference equality) if executed from a
 * second call to ComputeTable (with a smaller table).
 */
static char cmpTeamsByPtsGls(struct mstatus *a, struct mstatus *b)
{
  if (a->Pkte > b->Pkte) return 1;
  if (a->Pkte < b->Pkte) return -1;
  if (a->Tore + b->GTore > b->Tore + a->GTore) return 1;
  if (a->Tore + b->GTore < b->Tore + a->GTore) return -1;
  if (a->Tore > b->Tore) return 1;
  if (a->Tore < b->Tore) return -1;
  return 0;
}

static void AddMStatus(struct mstatus *table, Int8 n, Int8 n1, Int8 n2)
{
  Int8 pkte = 3;
  if (n>0) n--;

  table[n].Tore  += n1;
  table[n].GTore += n2;
  table[n].Spiele++;
  if (n1>n2) {
    table[n].Pkte+=pkte;
    table[n].Gew++;
  } else if (n1==n2) {
    table[n].Pkte++;
    table[n].Rem++;
  } else {
    table[n].Ver++;
  }
}

/*
 * compute table; 
 * table:     standings table
 * who:       bit field of teams to compute table for
 * recursive: flag if we are called recursive; happens while
 *            computing a sub table when points, goals and goal
 *            difference for some teams are equal
 * returns:   "final" flag.
 */
Int8 ComputeTable(struct mstatus *table, struct match *match,
		  Int8 who, Int8 recursive)
{
  Int16 i,j;
  Int8 final=1, 
    places[4] = { 0, }, onplace[4] = { 0, };

  // init variables;
  for (i=0; i<4; i++) {
    table[i].Tore = table[i].GTore = table[i].Pkte = table[i].Spiele
      = table[i].Gew = table[i].Rem = table[i].Ver = 0;
    table[i].Platz = 1;
  }

  // take every match of every team within "who"
  for (i=0; i<6; i++) {
    if (((1<<(match[i].team1-1)) & who) && ((1<<(match[i].team2-1)) & who) && 
	(match[i].erg1 != -1) && (match[i].erg2 != -1)) {
      AddMStatus(table,match[i].team1,match[i].erg1,match[i].erg2);
      AddMStatus(table,match[i].team2,match[i].erg2,match[i].erg1);
    } else {
      // there is at least one match missing for a final table.
      final = 0;
    }
  }

  // actually compute table for every team within "who"
  for (i=0; i<3; i++)
    for (j=i+1; j<4; j++)
      if ((1<<i & who) && ((1<<j & who))) {
	char sort = cmpTeamsByPtsGls(&table[i],&table[j]);
	if (sort<0)
	  table[i].Platz++;
	else if (sort>0)
	  table[j].Platz++;
      }

  // if we have all results, check if some teams are equally placed.
  // If so, compute a table with all matches of these teams. Required
  // by Article 28) of FIFA WC 2002 Reglement.
  if (final && !recursive) {
    for (i=0; i<4; i++) {
      places[table[i].Platz-1]++; // count teams on place 
      onplace[table[i].Platz-1] |= 1<<i; // remember team
    }
    for (i=0; i<4; i++)
      if (places[i]>1) { // at least two teams with equal place
	ComputeTable(Equal,match,onplace[i],1);
	for (j=0; j<4; j++) { // evaluate newly computed table
	  table[j].Platz += Equal[j].Platz-1;
	}
      }
  }
  return final;
}

/**
 * Form Compute Table Functions
 */

static void DrawTableItem(void* tableP, Int16 row, Int16 column, 
			  RectangleType *bounds)
{
  Int16 n;
  UInt16 x;
  Char buf[25];
  struct team *team;

  n = TblGetItemInt(tableP, row, column);
  x = bounds->topLeft.x;

  StrIToA(buf,n);

  switch (column) {
  case 0:
    StrCat(buf,".");
    break;
  case 1:
    team = GetTeamString((appPrefs.group-1)*4+n);
    MemMove(buf,team->name,StrLen(team->name)+1);
    break;
  case 3:
    StrCat(buf," :");
    x += bounds->extent.x - FntCharsWidth(buf,StrLen(buf)) - 1;
    break;
  }

  FntSetFont(0);
  WinEraseRectangle(bounds, 0);
  WinDrawChars(buf, StrLen(buf), x, bounds->topLeft.y);
}  

/*
 * fills in the values for the table and redraws it.
 */
static void InitTable(void)
{
  Int16 Platz, n, row=0;
 
  for (Platz=1; Platz<=4; Platz++) {
    for (n=0; n<4; n++) {
      if (MStatus[n].Platz == Platz) {
        TblSetItemInt(leagueTable, row, 0, Platz);
        TblSetItemInt(leagueTable, row, 1, n+1);
        TblSetItemInt(leagueTable, row, 2, MStatus[n].Spiele);
        TblSetItemInt(leagueTable, row, 3, MStatus[n].Tore);
        TblSetItemInt(leagueTable, row, 4, MStatus[n].GTore);
        TblSetItemInt(leagueTable, row, 5, MStatus[n].Pkte);
        row++;
      }
    }    
  }
}

/*
 * get the contents of the editable fields in the form and save them
 */
static void SaveMatchdayResults(void)
{
  Int16 i;
  char *e1, *e2;
  Int8 buf[2], rows;

  for (i=0; i<8; i++) {
    e1 = FldGetTextPtr(match[i].e1);
    e2 = FldGetTextPtr(match[i].e2);

    buf[0] = buf[1] = 255;

    if (e1 && StrLen(e1) && TxtCharIsDigit(*e1)
	&& e2 && StrLen(e2) && TxtCharIsDigit(*e2)) {
      buf[0] = StrAToI(e1);
      buf[1] = StrAToI(e2);
    }
    match[i].erg1 = buf[0];
    match[i].erg2 = buf[1];
  }

  if (appPrefs.group > 0) rows = 6;
  else rows = runden[appPrefs.round+1]-runden[appPrefs.round];
  WriteFixture(match, rows, appPrefs.group);
}

void ReadSetFixture(struct match* match, UInt8 group, Boolean setallfields)
{
  struct spiel Spiel;
  char *p;
  UInt8 i = 0, j = 0, recNo;
  for (p=ReadFixture(0, &Spiel),recNo=0; p; p=ReadFixture(p, &Spiel),recNo++) {
    if (Spiel.gruppe != group)
      continue;
    if (group == 0) {
      j++;
      if (j<=runden[appPrefs.round] || j>runden[appPrefs.round+1])
	continue;
    }
    match[i].datum = Spiel.datum;
    match[i].ort = Spiel.ort;
    match[i].erg1 = Spiel.e1;
    match[i].erg2 = Spiel.e2;
    match[i].team1 = Spiel.m1;
    match[i].team2 = Spiel.m2;
    match[i].recNo = recNo;
    if (setallfields) {
      if (group>0) {
	SetHandleString(match[i].m1, (group-1)*4 + Spiel.m1-1);
	SetHandleString(match[i].m2, (group-1)*4 + Spiel.m2-1);
	match[i].team1Final = Spiel.m1+(group-1)*4;
	match[i].team2Final = Spiel.m2+(group-1)*4;
      } else {
	match[i].team1Final = ComputeSetHandleString(match[i].m1, Spiel.m1);
	match[i].team2Final = ComputeSetHandleString(match[i].m2, Spiel.m2);
	SetHandleInt(match[i].num,recNo+1);
      }
      SetHandleInt(match[i].e1,Spiel.e1);
      SetHandleInt(match[i].e2,Spiel.e2);
    }
    i++;
  }
}

void InitFormMain(void)
{
  Int16 i, j, rows;
  FormType* frm = FrmGetActiveForm();

  selectedStadium = 0;
  selectedTeam = 0;

  if (appPrefs.group > 0) rows = 6;
  else rows = runden[appPrefs.round+1]-runden[appPrefs.round];

  /*
   * initialize table variables to default values;
   */
  leagueTable = Id2Ptr(IdFrmMain+100);

  for (i=0; i<8; i++) {
    match[i].m1 = Id2Ptr(IdFrmMain+20+i);
    match[i].m2 = Id2Ptr(IdFrmMain+30+i);
    match[i].e1 = Id2Ptr(IdFrmMain+40+i);
    match[i].e2 = Id2Ptr(IdFrmMain+50+i);
    match[i].num = Id2Ptr(IdFrmMain+60+i);
    FldSetUsable(match[i].m1, i<rows?true:false);
    FldSetUsable(match[i].m2, i<rows?true:false);
    FldSetUsable(match[i].e1, i<rows?true:false);
    FldSetUsable(match[i].e2, i<rows?true:false);
    FldSetUsable(match[i].num, i<rows?true:false);
    CtlSetUsable(Id2Ptr(IdFrmMain+10+i), i<rows?true:false);
  }

  ReadSetFixture(match,appPrefs.group,true);
  if (appPrefs.group>0) {
    FrmHideObject(frm, FrmGetObjectIndex(frm,IdFrmMain+3));
    FrmHideObject(frm, FrmGetObjectIndex(frm,IdFrmMain+4));
    for (i=0; i<4; i++) {
      TblSetRowSelectable(leagueTable, i, false);
      TblSetRowUsable(leagueTable, i, appPrefs.group>0?true:false);
      for (j=0; j<6; j++) {
	TblSetItemStyle(leagueTable, i, j, customTableItem);
	TblSetColumnUsable(leagueTable, j, true);
	TblSetCustomDrawProcedure(leagueTable, j, DrawTableItem);
      }
    }
    ComputeTable(MStatus,match,15,0); // 15 => all bits set, all teams
    InitTable();
  } else {
    FrmHideObject(frm, FrmGetObjectIndex(frm,IdFrmMain+2));
    FrmHideObject(frm, FrmGetObjectIndex(frm,IdFrmMain+7));
    FrmHideObject(frm, FrmGetObjectIndex(frm,IdFrmMain+8));
    FrmSetObjectPosition(frm, FrmGetObjectIndex(frm,IdFrmMain+100),0,159);
  }
  FrmCopyLabel(frm, IdFrmMain, 
	       LstGetSelectionText(Id2Ptr(IdFrmMain+1),appPrefs.group));
  FrmDrawForm(frm);
  SetRoundString(IdFrmMain+5,IdRoundString+(appPrefs.group>0 ? 0 : appPrefs.round+1));
}

Boolean SpieltagFormHandleEvent(EventPtr e)
{
  Boolean handled = false;
  Int16 item;
  ListPtr pLst;
    
  switch (e->eType) {

  case frmOpenEvent:
    InitFormMain();
    handled = true; 
    break;

  case menuEvent:

    switch(e->data.menu.itemID) {

    case IdMenuItem:
      FrmPopupForm(IdFrmAbout);
      break;

    case IdMenuItem+1:
      FrmHelp(IdFrmMain);
      break;

    case IdMenuItem+2:
      FrmPopupForm(IdFrmTimezone);
      break;

    case IdMenuItem+4:
      FrmPopupForm(IdFrmMatches);
      break;
    }
    handled = true;
    break;

  case ctlSelectEvent:
    switch (e->data.ctlEnter.controlID) {

    case IdFrmMain:
      pLst = Id2Ptr(IdFrmMain+1);
      LstSetSelection(pLst, appPrefs.group);
      LstMakeItemVisible(pLst, appPrefs.group);
      LstPopupList(pLst);
      if ( (item = LstGetSelection(pLst)) != appPrefs.group && item!=-1) {
	SaveMatchdayResults();
	appPrefs.group = item;
	FrmGotoForm(IdFrmMain);
      }
      handled = true;
      break;

    case IdFrmMain+2: /* Down-Arrow => Compute Table */
      SaveMatchdayResults();
      ComputeTable(MStatus,match,15,0); // 15 => all bits set, all teams
      InitTable();
      TblDrawTable(leagueTable);
      handled = true;
      break;

    case IdFrmMain+3: /* Left-Arrow  */
      if (appPrefs.round>0) { 
	SaveMatchdayResults();
	appPrefs.round--;
	FrmGotoForm(IdFrmMain);
      }
      handled = true;
      break;

    case IdFrmMain+4: /* Right-Arrow */
      if (appPrefs.round<3) {
	SaveMatchdayResults();
	appPrefs.round++;
	FrmGotoForm(IdFrmMain);
      }
      handled = true;
      break;

    case IdFrmMain+5: /* Stadium Button */
      if (selectedStadium>0)
	FrmPopupForm(IdFrmStadium);
      handled = true;
      break;

    default:
      /* Minus-Char => show place  and date of match */
      if (e->data.ctlEnter.controlID >= IdFrmMain+10 &&
	  e->data.ctlEnter.controlID <= IdFrmMain+17) {
	int index = e->data.ctlEnter.controlID - IdFrmMain - 10;
	selectedStadium = match[index].ort;
	SetOrtString(IdFrmMain + 5, match[index].ort-1);
	SetDatumString(IdFrmMain + 6, match[index].datum);
	handled = true;
      }
    }
    break;

  case fldEnterEvent:
    item = e->data.fldEnter.fieldID-IdFrmMain-20;
    if (item>=0 && item<8) {
      handled = true;
      selectedTeam = match[item].team1Final;
    } else if (item>=10 && item<18) {
      handled = true;
      item -= 10;
      selectedTeam = match[item].team2Final;
    }
    if (handled && selectedTeam>0 && selectedTeam<=32)
      FrmPopupForm(IdFrmTeams);
    break;

  case keyDownEvent:
    switch (e->data.keyDown.chr) {
    case pageUpChr:
      SaveMatchdayResults();
      if (appPrefs.group == 0) {
	if (appPrefs.round>0) {
	  appPrefs.round--;
	  FrmGotoForm(IdFrmMain);
	}
      } else {
	if (appPrefs.group>1) {
	  appPrefs.group--;
	  FrmGotoForm(IdFrmMain);
	}
      }
      handled = true;
      break;
    case pageDownChr:
      SaveMatchdayResults();
      if (appPrefs.group == 0) {
	if (appPrefs.round<3) {
	  appPrefs.round++;
	  FrmGotoForm(IdFrmMain);
	}
      } else {
	if (appPrefs.group<8) {
	  appPrefs.group++;
	  FrmGotoForm(IdFrmMain);
	}
      }
      handled = true;
      break;
    }
    break;

  default:
    break;
  }

  return handled;
}

/**
 * About Form
 */
Boolean AboutFormHandleEvent(EventPtr e)
{
  Boolean handled;
  
  handled = false;
  switch (e->eType) {

  case frmOpenEvent:
    FrmDrawForm(FrmGetActiveForm());
    handled = true; 
    break;

  case ctlSelectEvent:
    if (e->data.ctlSelect.controlID == IdFrmAbout) {
      FrmReturnToForm(0);
      handled = true; 
    }
    break;

  default:
    break;
  }

  return handled;
}

/**
 * Timezone Form
 */
Boolean TimezoneFormHandleEvent(EventPtr e)
{
  static UInt8 element;
  static Int32 newGmtOff, minutes;
  Boolean handled;
  FormPtr frm;
  
  handled = false;
  switch (e->eType) {

  case frmOpenEvent:
    element = 0;
    // just to be sure, that the date really is correctly "aligned"
    newGmtOff = (appPrefs.gmtOff/1800L)*1800L;
    minutes = (newGmtOff%3600L)/60L;
    frm = FrmGetActiveForm();
    FrmSetControlGroupSelection(frm, 1, IdFrmTimezone+2);
    FrmDrawForm(frm);
    SetTimezoneString(IdFrmTimezone+2,newGmtOff/3600L,"%+2d");
    SetTimezoneString(IdFrmTimezone+3,minutes,"%02d");
    handled = true; 
    break;

  case ctlSelectEvent:
    switch (e->data.ctlSelect.controlID) {
    case IdFrmTimezone:
      appPrefs.gmtOff = newGmtOff;
    case IdFrmTimezone+1:
      FrmReturnToForm(0);
      handled = true; 
      break;

    case IdFrmTimezone+2:
      element = 0;
      handled = true;
      break;
    case IdFrmTimezone+3:
      element = 1;
      handled = true;
      break;

    case IdFrmTimezone+4:
      handled = true;
      /* according to a time zone map there really _is_ a time zone
       * that's off by +1400...
       */
      if (element == 0) {
	if (newGmtOff<=13L*3600L) {
	  newGmtOff += 3600L;
	  SetTimezoneString(IdFrmTimezone+2,newGmtOff/3600L,"%+2d");
	}
	break;
      }
    case IdFrmTimezone+5:
      handled = true;
      if (element == 1) { // doesn't matter if down or up...
	newGmtOff -= 60L*minutes;
	minutes = (30-minutes);
	newGmtOff += 60L*minutes;
	SetTimezoneString(IdFrmTimezone+3,minutes,"%02d");
	break;
      }
      if (element == 0 && newGmtOff>=-11L*3600L) {
	newGmtOff -= 3600L;
	SetTimezoneString(IdFrmTimezone+2,newGmtOff/3600L,"%+2d");
	break;
      }
    }
    break;

  default:
    break;
  }

  return handled;
}

